import pytest
from django.contrib.auth import get_user_model
from rest_framework.test import APIRequestFactory

from .. import custom_exceptions
from ..authentication import LearngualAuthentication

factory = APIRequestFactory()
User = get_user_model()


def test_authentication_service_key(settings):

    request = factory.get("/api/my-endpoint/")

    # Add the custom header to the request

    request.META["HTTP_SERVICE_KEY"] = settings.LEARNGUAL_SERVICE_API_KEY

    # Authenticate the request

    # Create an instance of the custom authentication class
    auth = LearngualAuthentication()

    # Authenticate the request using the custom authentication class
    authenticated = auth.authenticate(request)

    # Assert that authentication is successful
    assert getattr(authenticated[0], "is_service", None)
    assert authenticated[1] == settings.LEARNGUAL_SERVICE_API_KEY


@pytest.mark.django_db
def test_authentication_invalid_service_key(settings):

    request = factory.get("/api/my-endpoint/")

    # Add the custom header to the request

    request.META["HTTP_SERVICE_KEY"] = "hshsggfaas"

    # Authenticate the request

    # Create an instance of the custom authentication class
    auth = LearngualAuthentication()

    try:
        # Authenticate the request using the custom authentication class
        authenticated = auth.authenticate(request)
        assert authenticated
    except custom_exceptions.UnAuthenticated:
        assert True



from django.db.models import PositiveIntegerField, Subquery


class SubqueryCount(Subquery):
    """
    from django.db.models import OuterRef

    weapons = Weapon.objects.filter(unit__player_id=OuterRef('id'))
    units = Unit.objects.filter(player_id=OuterRef('id'))

    qs = Player.objects.annotate(weapon_count=SubqueryCount(weapons),
                                rarity_sum=SubquerySum(units, 'rarity'))
    """

    # Custom Count function to just perform simple count on any queryset without grouping.
    # https://stackoverflow.com/a/47371514/1164966
    template = "(SELECT count(*) FROM (%(subquery)s) _count)"
    output_field = PositiveIntegerField()


class SubqueryAggregate(Subquery):
    # https://code.djangoproject.com/ticket/10060
    template = '(SELECT %(function)s(_agg."%(column)s") FROM (%(subquery)s) _agg)'

    def __init__(self, queryset, column, output_field=None, **extra):
        if not output_field:
            # infer output_field from field type
            output_field = queryset.model._meta.get_field(column)
        super().__init__(
            queryset, output_field, column=column, function=self.function, **extra
        )


class SubquerySum(SubqueryAggregate):
    function = "SUM"



from typing import Literal

from pydantic import AnyHttpUrl, Field

from .interface import BaseModel, BaseTypeModel


class SocialMediaLinksSchema(BaseModel):
    facebook: AnyHttpUrl | None = Field(None)
    instagram: AnyHttpUrl | None = Field(None)
    linkedin: AnyHttpUrl | None = Field(None)
    twitter: AnyHttpUrl | None = Field(None)
    website: AnyHttpUrl | None = Field(None)


"""
Account settings link
"""


class AccountSettingsSchema:
    class Base(BaseTypeModel):
        type: Literal["DEFAULT"] = "DEFAULT"
        social_links: SocialMediaLinksSchema = Field(
            default_factory=SocialMediaLinksSchema
        )

    class EnterpriseSettings(Base, BaseTypeModel):

        type: Literal["ENTERPRISE"] = "ENTERPRISE"
        who_can_view_your_profile: Literal[
            "Enterprise members only",
            "Enterprise members and public viewers",
            "Public viewers",
        ] = Field("Enterprise members only")
        show_statistics: bool = Field(False)
        show_available_course_languages: bool = Field(False)
        show_social_media_links: bool = Field(False)
        show_courses: bool = Field(False)

    class InstructorSettings(Base):
        type: Literal["INSTRUCTOR"] = "INSTRUCTOR"
        who_can_view_your_profile: Literal[
            "Students only",
            "Instructors only",
            "Students and instructors",
            "All enterprise",
        ] = Field("Friends")
        show_statistics: bool = Field(True)
        show_available_course_languages: bool = Field(True)
        show_social_media_links: bool = Field(True)
        show_courses: bool = Field(True)
        show_affiliates: bool = Field(True)

    class PersonalSettings(Base):
        who_can_view_your_profile: Literal[
            "Friends",
            "All affiliated accounts",
            "Public",
            "Students and enterprise",
            "Students and instructors",
        ] = Field("Friends")
        show_evaluations: bool = Field(True)
        show_achievements: bool = Field(True)
        show_affiliate: bool = Field(True)
        show_courses: bool = Field(True)
        exercises_per_week: int = Field(0)

        type: Literal["PERSONAL"] = "PERSONAL"

    class DeveloperSettings(Base):
        type: Literal["DEVELOPER"] = "DEVELOPER"



import importlib
import logging
import os
import urllib.parse
from logging import getLogger

import requests
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.cache import cache
from django.http.request import HttpRequest
from django.utils import timezone, translation
from rest_framework.request import Request
from rest_framework_simplejwt import authentication
from rest_framework_simplejwt.exceptions import InvalidToken

from . import custom_exceptions
from .translator import translate as _
from .utils import decode_jwt_access_token, get_language_code

logger = getLogger(__file__)

User = get_user_model()

LEARNGUAL_SERVICE_API_KEY = getattr(
    settings, "LEARNGUAL_SERVICE_API_KEY", None
) or os.getenv("LEARNGUAL_SERVICE_API_KEY", None)

"""
Configure authentication class
REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": (
        ...
        "iam_service.users.authentication.LearngualAuthentication",
        ...
    ),
    ...
}

LEARNGUAL_AUTH_RETRIEVE_URL=<auth server url to authenticate if user is login>
LEARNGUAL_AUTH_TEST_MODE=<True or False used for testing provide dummy data to authenticate user>
LEARNGUAL_AUTH_GET_USER=<dot path to the callable >

Data = {
    "account": {
        "id": "84bcaf2972",
        "cover_photo": None,
        "profile_photo": None,
        "type": "PERSONNAL",
        "metadata": {},
        "created_at": "2023-01-13T16:33:52.084540Z",
        "updated_at": "2023-01-13T16:33:52.084576Z"
    },
    "email": "Bulah53@gmail.com",
    "first_name": "Caitlyn",
    "id": "40e0e7013f",
    "last_name": "Marquardt",
    "registration_step": "REGISTRATION_COMPLETED",
    "username": "Eloisa.Senger42"
}

def get_user(data:Data):

    print("data in get user")
    return get_user_model().objects.get_or_create()

"""


user_test_data = {
    "account": {
        "id": "84bcaf2972",
        "cover_photo": None,
        "profile_photo": None,
        "type": "PERSONAL",
        "metadata": {},
        "created_at": "2023-01-13T16:33:52.084540Z",
        "updated_at": "2023-01-13T16:33:52.084576Z",
    },
    "email": "Bulah53@gmail.com",
    "first_name": "Caitlyn",
    "id": "40e0e7013f",
    "last_name": "Marquardt",
    "registration_step": "REGISTRATION_COMPLETED",
    "username": "Eloisa.Senger42",
}


def load_callable(path: str) -> object | None:
    paths = path.split(".")
    modules = importlib.import_module(".".join(paths[:-1]))
    result = getattr(modules, paths[-1], None)
    if not result:
        logger.warning("Module does no exists. path: %s", path)
    return result


LEARNGUAL_AUTH_RETRIEVE_URL = getattr(settings, "LEARNGUAL_AUTH_RETRIEVE_URL", None)
LEARNGUAL_AUTH_GET_USER = getattr(settings, "LEARNGUAL_AUTH_GET_USER", None)
LEARNGUAL_AUTH_ACCOUNT_MODEL_PATH = getattr(
    settings, "LEARNGUAL_AUTH_ACCOUNT_MODEL_PATH", None
)


assert (
    LEARNGUAL_AUTH_RETRIEVE_URL
), "LEARNGUAL_AUTH_RETRIEVE_URL must be provided in the settings."
assert (
    LEARNGUAL_AUTH_GET_USER
), "LEARNGUAL_AUTH_GET_USER must be provided in the settings."


get_user: callable = load_callable(LEARNGUAL_AUTH_GET_USER)
assert get_user, "No callable exists in path:%s" % LEARNGUAL_AUTH_GET_USER


class AuthenticationBaseHelper:
    @classmethod
    def get_http_headers(cls, request: Request | HttpRequest):
        headers = {
            key.lower().lstrip("http_"): value
            for key, value in request.META.items()
            if key.lower().startswith("http_")
        }
        return headers

    @classmethod
    def get_query_str(cls, request: Request | HttpRequest):
        return "?{query}".format(query=request.META.get("QUERY_STRING", ""))

    def authenticate_service(self, request: Request | HttpRequest):
        headers = self.get_header(request)

        if service_key := headers.get("service_key"):
            logger.info("authentication send with service account")
            if service_key == LEARNGUAL_SERVICE_API_KEY:
                service_user = User(id="service", username="service")
                service_user.is_service = True
                return service_user, service_key
            logger.warning("invalid service key")
            msg = _("Invalid service key")
            raise custom_exceptions.UnAuthenticated(msg)

    def get_user(self, validated_token, headers):
        """
        Attempts to find and return a user using the given validated token.
        """
        try:
            user_id = validated_token[authentication.api_settings.USER_ID_CLAIM]
        except KeyError:
            raise InvalidToken(_("Token contained no recognizable user identification"))

        res_data = {}
        res_data["id"] = user_id
        if account_id := headers.get("account_id"):
            res_data["account"] = {"id": account_id}
        user = get_user(res_data)
        account = getattr(user, "account", None)
        if account and account.owner != user:
            logger.warning(f"{account = }, {account.owner = }, request user {user}")
            raise custom_exceptions.UnAuthenticated(_("Invalid account provided"))
        return user

    def get_raw_token(self, header):
        """
        Extracts an unvalidated JSON web token from the given "Authorization"
        header value.
        """
        parts = header.split()

        if len(parts) == 0:
            # Empty AUTHORIZATION header sent
            return None

        if str(parts[0]).lower().strip() != "bearer":
            # Assume the header does not contain a JSON web token
            return None

        if len(parts) != 2:
            raise custom_exceptions.UnAuthenticated(
                _("Authorization header must contain two space-delimited values"),
                code="bad_authorization_header",
            )

        return parts[1]

    def append_bearer(self, token: str):
        if (
            token
            and isinstance(token, str)
            and not token.lower().strip().startswith("bearer")
        ):
            return "Bearer " + token
        return token

    def get_header(self, request: Request | HttpRequest):
        """
        Extracts the header containing the JSON web token from the given
        request.
        """
        account_id = str(
            request.META.get("HTTP_ACCOUNT", "")
            or request.META.get("HTTP_X_ACCOUNT", "")
            or request.COOKIES.get("x-account", "")
            or request.COOKIES.get("account", "")
            or request.COOKIES.get("_account", "")
            or request.GET.get("x-account")
            or request.GET.get("_account")
            or ""
        )
        api_key = str(
            request.META.get("HTTP_API_KEY", "")
            or request.META.get("HTTP_X_API_KEY", "")
            or request.COOKIES.get("x-api-key", "")
            or request.COOKIES.get("api-key", "")
            or request.GET.get("x-api-key")
            or request.GET.get("api-key")
            or ""
        )
        service_key = str(
            request.META.get("HTTP_SERVICE_KEY", "")
            or request.META.get("HTTP_X_SERVICE_KEY", "")
            or request.COOKIES.get("x-service-key", "")
            or request.COOKIES.get("service-key", "")
            or request.GET.get("service-key", "")
            or request.GET.get("_service-key", "")
            or ""
        )
        authorization = str(
            request.META.get("HTTP_AUTHORIZATION", "")
            or request.COOKIES.get("x-authorization", "")
            or request.COOKIES.get("authorization", "")
            or request.COOKIES.get("x-token", "")
            or request.COOKIES.get("token", "")
            or request.GET.get("token", "")
            or request.GET.get("_token", "")
            or ""
        )

        return dict(
            account_id=urllib.parse.unquote(account_id or ""),
            api_key=urllib.parse.unquote(api_key or ""),
            authorization=urllib.parse.unquote(self.append_bearer(authorization or "")),
            service_key=urllib.parse.unquote(service_key or ""),
        )


class LearngualAuthentication(
    AuthenticationBaseHelper, authentication.JWTAuthentication
):
    """
    An authentication plugin that authenticates requests through a JSON web
    token provided in a request header.
    """

    def authenticate_api_key(self, request: Request | HttpRequest):
        header = self.get_header(request)
        if not (api_key := header.get("api_key")):
            return
        headers = self.get_http_headers(request)
        query_str = self.get_query_str(request)
        res = requests.get(LEARNGUAL_AUTH_RETRIEVE_URL + query_str, headers=headers)
        if not res.ok:
            logging.warning(
                "request to IAM service did not go through: %s", res.content
            )
            if res.status_code == 401:
                raise custom_exceptions.UnAuthenticated(_("Authentication required"))
            raise custom_exceptions.UnAuthenticated(_("Service is down."))
        res_data = res.json()
        if not (user := get_user(res_data)):
            return
        self.set_language(user, request)
        return (user, api_key)

    def authenticate(self, request: Request | HttpRequest):
        if is_service := self.authenticate_service(request):
            return is_service
        if is_api := self.authenticate_api_key(request):
            return is_api
        header = self.get_header(request)
        if not (authorization := header.get("authorization")):
            logger.warning("no authorization")
            return
        raw_token = self.get_raw_token(authorization)
        if raw_token is None:
            return
        try:
            validated_token = self.get_validated_token(raw_token)
        except InvalidToken as e:
            raise custom_exceptions.UnAuthorized(*e.args)
        data = decode_jwt_access_token(str(validated_token))
        jti = data.get("jti")
        if jti and cache.get(jti):
            raise custom_exceptions.UnAuthorized("token blacklisted")

        if not (user := self.get_user(validated_token, header)):
            return

        if user and user.token_blacklist_at:
            created_at: str = data.get("created_at")
            if (
                not created_at
                or timezone.datetime.fromisoformat(created_at.rstrip("zZ"))
                <= user.token_blacklist_at
            ):
                raise custom_exceptions.UnAuthorized("token blacklisted")
        self.set_language(user, request)
        return user, validated_token

    def set_language(self, user, request: Request | HttpRequest):
        if user and not request.GET.get("_lang"):
            if account := getattr(user, "account", None):
                translation.activate(get_language_code(account.language))
            else:
                translation.activate(get_language_code(user.language))



import importlib
import json
import os
from collections import defaultdict
from logging import getLogger

import gspread
from django.conf import settings
from django.core.cache import cache
from django.utils import timezone, translation
from google.oauth2.service_account import Credentials
from rest_framework.serializers import ValidationError

from .cache_models import CacheTranslation
from .enums import LanguageCodeType

logger = getLogger(__file__)


def load_callable(path: str) -> object | None:
    paths = path.split(".")
    modules = importlib.import_module(".".join(paths[:-1]))
    result = getattr(modules, paths[-1], None)
    if not result:
        logger.warning("Module does no exists. path: %s", path)
    return result


"""
from iam_service.learngual.translator import Translator
t = self = Translator(target_language="KO")
t.log_queue
"""


class Translator:
    key = "Translator-datav2"
    queue_key = "Translator-queue"

    refresh_duration = timezone.timedelta(minutes=1).total_seconds()

    def __init__(
        self,
        sheet_id=None,
        sheet_name=None,
        celery_app_path: str | None = None,
        target_language: str = None,
    ):
        if not target_language:
            target_language = translation.get_language() or "EN"
        target_language = target_language.upper()
        cache.delete_many(["Translator-data", "Translator-datav2", "Translator-queue"])
        self.sheet_id = (
            sheet_id
            or getattr(settings, "LEARNGUAL_TRANSLATE_SHEET_ID", None)
            or os.getenv("LEARNGUAL_TRANSLATE_SHEET_ID")
        )
        self.sheet_name = (
            sheet_name
            or getattr(settings, "LEARNGUAL_TRANSLATE_SHEET_NAME", None)
            or os.getenv("LEARNGUAL_TRANSLATE_SHEET_NAME")
        )

        self.celery_app_path = (
            celery_app_path
            or getattr(settings, "LEARNGUAL_CELERY_APP_PATH", None)
            or os.getenv("LEARNGUAL_CELERY_APP_PATH")
        )

        assert (
            self.sheet_id
        ), "`LEARNGUAL_TRANSLATE_SHEET_ID` must be set in enviroment variable or Django setting"
        assert (
            self.sheet_name
        ), "`LEARNGUAL_TRANSLATE_SHEET_NAME` must be set in enviroment variable or Django setting"

        scopes = ["https://www.googleapis.com/auth/spreadsheets"]
        creds = json.loads(
            getattr(settings, "LEARNGUAL_GOOGLE_BOT_GRED", None)
            or os.getenv("LEARNGUAL_GOOGLE_BOT_GRED")
            or "{}"
        )
        credentials = Credentials.from_service_account_info(creds, scopes=scopes)
        self.gc = gspread.authorize(credentials)

        self.target_language = target_language

    @property
    def translations(self):
        return self.get_cache_model().get_default_cache_model().translation

    @property
    def headers(self):
        return self.get_cache_model().get_default_cache_model().headers

    @property
    def celery_app(self):
        return load_callable(self.celery_app_path)

    def remove_duplicate_rows(self, column_name="en"):
        # Authenticate and open the Google Sheet

        sheet = self.gc.open_by_key(self.sheet_id)
        worksheet = sheet.worksheet(self.sheet_name)

        # Get all data from the sheet
        data = worksheet.get_all_values()

        # Extract header and data rows
        header = data[0]
        rows = data[1:]

        # Find the index of the target column
        try:
            if column_name.lower() in header:
                column_index = header.index(column_name.lower())
            elif column_name.upper() in header:
                column_index = header.index(column_name.upper())
            else:
                column_index = header.index(column_name)
        except ValueError:
            raise ValueError(f"Column '{column_name}' not found in the sheet.")

        # Use a dictionary to track the last occurrence of each value in the target column
        last_occurrence = defaultdict(int)

        # Iterate through the rows to find the last occurrence of each value
        for i, row in enumerate(rows):
            value = row[column_index]
            last_occurrence[value] = i + 1  # +1 because of header

        # Build a list of row indices to keep
        rows_to_keep = set(last_occurrence.values())

        # Delete rows that are not in the rows_to_keep set
        rows_to_delete = [
            i + 2 for i in range(len(rows)) if (i + 1) not in rows_to_keep
        ]
        rows_to_delete.reverse()  # Delete from bottom to top to avoid shifting issues

        for row_index in rows_to_delete:
            worksheet.delete_rows(row_index)

        logger.info(
            f"Removed {len(rows_to_delete)} duplicate rows from column '{column_name}'."
        )

    def get_cache_model(self) -> CacheTranslation:
        return CacheTranslation

    def update_translation(self, headers, translations):

        self.get_cache_model().get_default_cache_model().update_translations(
            translations=translations, headers=headers
        )

    def update_queue(self, *texts: str):
        return self.get_cache_model().get_default_cache_model().add_to_queue(*texts)

    def overwrite_queue(self, texts: list):
        model = self.get_cache_model().get_default_cache_model()
        model.queue = texts

    def force_load_translations(self):
        logger.info("Initiate refresh tranlsations")
        try:
            self.remove_duplicate_rows()
        except Exception:
            ...
        sheet = self.gc.open_by_key(self.sheet_id)
        worksheet = sheet.worksheet(self.sheet_name)
        data = worksheet.get_all_values()

        headers = [x.upper() for x in data[0]]
        translations = {}

        if not all([x in headers for x in ["EN", "KO"]]):
            raise ValidationError("No translation for EN or KO.")

        for row in data[1:]:
            en_value = row[0].strip()
            translations[en_value] = {
                headers[i]: row[i] for i in range(1, len(headers))
            }
        logger.info(f"Update Cache {len(translations.keys())}")
        self.update_translation(headers=headers, translations=translations)

        return translations

    def get_language_code(self, language: str) -> str:
        languages = {
            key.strip().upper(): value.strip().upper()
            for key, value in LanguageCodeType.dict_name_key().items()
        }
        if not language:
            language = "EN"
        language = language.strip().upper()
        return languages.get(language, language)

    @property
    def log_queue(self):
        return self.get_cache_model().get_default_cache_model().queue

    def get_language_list(self):
        """{
            "EN":[],
            "KO":[],
        }

        Returns:
            _type_: _description_
        """
        data = {key.upper(): [] for key in self.headers}
        data["EN"] = []
        for en_value, value in self.translations.items():
            data["EN"].append(en_value)
            for key, micro in value.items():
                if micro and key in data:
                    data[key].append(micro)

        return data

    def log_microcopy(self, text: str):
        logger.info(f"Log microcopy: {text}")
        if text not in self.translations and text:
            translations = self.translations
            headers = self.headers
            translations[text] = {headers[i]: "" for i in range(1, len(headers))}
            self.update_translation(
                **{"headers": headers, "translations": translations}
            )

            if text not in self.log_queue:
                self.update_queue(text)
                logger.info(f"Add microcopy '{text}' to log queue.")

    def log_microcopy_job(self):
        queue_list = self.log_queue

        sheet = self.gc.open_by_key(self.sheet_id)
        worksheet = sheet.worksheet(self.sheet_name)
        data = worksheet.get_all_values()

        translations = {}
        headers = [x.upper() for x in data[0]]

        if not all([x in headers for x in ["EN", "KO"]]):
            logger.info("EN and KO column not found")
            raise ValidationError("No translation for EN and KO")

        for row in data[1:]:
            en_value = row[0].strip()
            translations[en_value] = {
                headers[i]: row[i] for i in range(1, len(headers))
            }

        queue_list = [x for x in self.log_queue if x not in translations]
        batch_to_update = queue_list[:60]
        batch_to_keep = queue_list[61:]
        if self.log_queue:
            self.overwrite_queue(batch_to_keep)
        if batch_to_update:
            logger.info(f"Added {len(batch_to_update)} microcopies to queue to sheet")
            start_index = 2
            worksheet.insert_rows([[x] for x in batch_to_update], row=start_index)

        self.force_load_translations()

    def force_add_language_column(self, language: str):
        if self.headers and language not in self.headers:
            sheet = self.gc.open_by_key(self.sheet_id)
            worksheet = sheet.worksheet(self.sheet_name)
            # Add new column with target language header
            col_count = len(worksheet.row_values(1))
            worksheet.update_cell(1, col_count + 1, language.upper())
            logger.info(f"Added new language column: {language}")

    def add_language_column(self, language: str):
        if self.headers and language not in self.headers:
            headers: list = self.headers
            headers.append(language.upper())
            self.update_translation(
                **{"headers": headers, "translations": self.translations}
            )

            self.force_add_language_column(language=language)

    def translate(self, text: str, target_language: str):
        target_language = self.get_language_code(target_language)
        if target_language.upper() not in self.headers:
            if not self.headers:
                self.celery_app.send_task(
                    "iam.accounts.translate_load_translations",
                    routing_key="iam.accounts.translate_load_translations",
                )
            # else:
            #     self.add_language_column(target_language.upper())
            logger.info(f"Language: {target_language} does not exists in sheet")
            return text
        stripped_text = text.strip()
        translated_dict = self.translations.get(stripped_text)
        if not translated_dict:
            if translated_dict is None:
                if text not in (
                    self.get_language_list()[self.get_language_code(target_language)]
                    or []
                ):
                    self.log_microcopy(stripped_text)
            return stripped_text

        translated_text = translated_dict.get(target_language, text)
        if not translated_text:
            return stripped_text

        return translated_text

    def get_translation(
        self, text, target_language: str | None = None, **kwargs
    ) -> str:
        target_language = target_language or self.target_language
        assert target_language, "target_language is required."
        result = self.translate(text, target_language)
        if kwargs:
            result = self.render(result, **kwargs)

        return result

    def render(self, text: str, **kwargs):
        if not text:
            return text
        try:
            return text.format(**kwargs)
        except KeyError as e:
            raise ValidationError(
                {"detail": f"Imcomplete kwargs for {kwargs} for text {text}"}
            ) from e


def translate(text, target_language: str | None = None, **kwargs) -> str:
    """Function to translate text
    use _(text)

    Args:
        text (_type_): _description_
        target_language (str | None, optional): _description_. Defaults to None.

    Returns:
        str: _description_
    """
    return Translator().get_translation(text, target_language=target_language, **kwargs)



import json

from django.db import models
from pydantic import BaseModel
from pydantic import ValidationError as PydanticValidationError
from rest_framework import serializers

from .interface import BaseTypeModel
from .logger import logger


class PydanticModelFieldEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, BaseModel):
            return obj.model_dump(mode="json")
        elif isinstance(obj, list) and isinstance(obj[0], BaseModel):
            data: list[BaseModel] = obj
            return [model.model_dump(mode="json") for model in data]
        else:
            return super().default(obj)


class PydanticModelField(models.JSONField):
    """Usage

    data = PydanticModelField(pydantic_model=AnswerModel) \n
    data = PydanticModelField(pydantic_model=[AnswerModel]) \n
    data = PydanticModelField(pydantic_model={"word_play":AnswerModel,"essay":EssayAnswer})
    """

    def __init__(
        self,
        pydantic_model: BaseModel | tuple[BaseModel] | dict[str, BaseModel] = None,
        null=True,
        blank=True,
        *args,
        **kwargs,
    ):
        """Pydantic Model field

        Args:
            pydantic_model (BaseModel | Tuple[BaseModel] | dict[str,BaseModel], optional): _description_. Defaults to None. # noqa

        Raises:
            ValueError: _description_
            ValueError: _description_
            ValueError: _description_
        """

        if pydantic_model:
            if isinstance(pydantic_model, (list, tuple)):
                if not pydantic_model:
                    raise ValueError("pydantic_model list cannot be empty")

                for model_class in pydantic_model:
                    if not issubclass(model_class, (BaseModel, BaseTypeModel)):
                        raise ValueError(
                            "All elements in the tuple/list must be subclasses of BaseModel"
                        )
            elif isinstance(pydantic_model, dict):
                for model_class in pydantic_model.values():
                    if not issubclass(model_class, BaseTypeModel):
                        raise ValueError(
                            "All values in the dictionary must be subclasses of BaseTypeModel"
                        )
            elif not issubclass(pydantic_model, (BaseModel, BaseTypeModel)):
                raise ValueError("pydantic_model must be a subclass of BaseModel")
        self.pydantic_model: BaseModel | BaseTypeModel | None | tuple[
            BaseModel | BaseTypeModel
        ] | dict[str, BaseModel | BaseTypeModel] = pydantic_model

        kwargs["encoder"] = kwargs.get("encoder", PydanticModelFieldEncoder)
        super().__init__(null=null, blank=blank, *args, **kwargs)

    def to_python(self, value):
        if isinstance(value, str):
            try:
                value = json.loads(value)
            except (TypeError, ValueError):
                pass

        if value and self.pydantic_model:
            try:
                if isinstance(self.pydantic_model, (list, tuple)):
                    data = []
                    ModelClass = self.pydantic_model[0]
                    value: list[dict]
                    for x in value:
                        data.append(ModelClass(**x))
                    return data
                elif isinstance(self.pydantic_model, dict):
                    value: dict
                    ModelClass = self.pydantic_model.get(value.get("type"))
                    return ModelClass(**value.get("data"))
                elif issubclass(self.pydantic_model, (BaseModel, BaseTypeModel)):
                    return self.pydantic_model(**value)
                else:
                    raise ValueError("Invalid data")
            except Exception:
                logger.exception("invalid data")
                return
                # raise ValidationError(f"Invalid data for {self.pydantic_model}: {e}")

        return value

    def from_db_value(self, value, expression, connection):
        return self.to_python(value)

    def get_prep_value(self, value: BaseModel | BaseTypeModel | list[BaseModel] | None):
        if value is None:
            return [] if isinstance(self.pydantic_model, (list, tuple)) else {}
        data = {}

        if not self.pydantic_model:
            return

        if isinstance(self.pydantic_model, (list, tuple)):
            if not isinstance(value, (list, tuple)):
                raise ValueError("Value must be a list or tuple")
            data = []
            ModelClass = self.pydantic_model[0]
            for model_instance in value:
                model_instance: BaseModel | BaseTypeModel
                if not isinstance(model_instance, ModelClass):
                    raise ValueError("Value must be a list %s" % str(ModelClass))
                data.append(model_instance.model_dump(mode="json"))
        elif isinstance(self.pydantic_model, dict):
            if not isinstance(value, BaseTypeModel):
                raise ValueError("Value must be an instance of BaseTypeModel")
            _type = getattr(value, "type", None)
            if not _type:
                raise ValueError(
                    "%s must have a field `type`" % value.__class__.__name__
                )
            model_type = _type
            ModelClass = self.pydantic_model.get(model_type)
            if ModelClass is None:
                raise ValueError("Invalid value.type")

            if not isinstance(value, ModelClass):
                raise ValueError("Value must be a list %s" % str(ModelClass))
            data = {"type": model_type, "data": value.model_dump(mode="json")}
        elif issubclass(self.pydantic_model, (BaseModel, BaseTypeModel)):
            if not isinstance(value, BaseModel):
                if not value:
                    return value
                raise ValueError(
                    f"Value must be an instance of BaseModel {value = } {self.pydantic_model = }"
                )
            return value.model_dump_json()
        else:
            raise ValueError("Invalid data")
        return json.dumps(data)

    def value_to_string(self, obj):
        value = self.value_from_object(obj)
        return self.get_prep_value(value)


class PydanticModelSerializerField(serializers.JSONField):
    """
    Example:
        class TestModelRetrieve(serializers.ModelSerializer):
            data = PydanticModelSerializerField(
                pydantic_model=modelsv2.TestModel.data.field.pydantic_model
            )\n
            list_data = PydanticModelSerializerField(
                pydantic_model=modelsv2.TestModel.list_data.field.pydantic_model
            )\n
            type_data = PydanticModelSerializerField(
                pydantic_model=modelsv2.TestModel.type_data.field.pydantic_model
            )\n

            class Meta:
                model = modelsv2.TestModel\n
                fields = [
                    "data",
                    "list_data",
                    "type_data",
                ]


        data with different type will be

        {
            "type":"WORD_PLAY",
            "data":{
                "key1":"value1",
            }
        }

    Args:
        serializers (_type_): _description_
    """

    def __init__(self, pydantic_model=None, *args, **kwargs):
        self.pydantic_model = pydantic_model
        super().__init__(*args, **kwargs)

    def to_internal_value(self, data: dict):
        try:
            if self.pydantic_model:
                if isinstance(self.pydantic_model, (list, tuple)):
                    deserialized_data = []
                    ModelClass = self.pydantic_model[0]
                    for item in data:
                        deserialized_data.append(ModelClass(**item))
                    return deserialized_data
                elif isinstance(self.pydantic_model, dict):
                    ModelClass = self.pydantic_model.get(data.get("type"))
                    if not ModelClass:
                        raise serializers.ValidationError(
                            f"Invalid type; {data.get('type')}"
                        )
                    return ModelClass(**data)
                elif issubclass(self.pydantic_model, (BaseModel, BaseTypeModel)):

                    return self.pydantic_model(**data)
                else:
                    raise serializers.ValidationError("Invalid data")
            return data
        except PydanticValidationError as e:
            raise serializers.ValidationError(e.json())

    def to_representation(self, value) -> str:
        data = PydanticModelField(self.pydantic_model).get_prep_value(value)
        if isinstance(data, str):
            data = json.loads(data)
        if (
            data
            and isinstance(data, dict)
            and sorted(list(data.keys())) == sorted(["data", "type"])
            and data.get("type")
            and data.get("type") == data.get("data", {}).get("type")
        ):
            data = data.get("data")
        return data

    def _format_validation_error(self, error):
        if isinstance(error, PydanticValidationError):
            error_messages = []

            for error_obj in error.errors():
                if isinstance(error_obj, dict):
                    for field, field_errors in error_obj.items():
                        for sub_error in field_errors:
                            error_messages.append(
                                {
                                    "field": field,
                                    "message": str(sub_error),
                                }
                            )
                else:
                    error_messages.append({"field": "", "message": str(error_obj)})

            return error_messages
        return str(error)



import json
import logging
import os


class JsonFormatter(logging.Formatter):
    """
    Custom formatter to output logs in JSON format, including logger name, extra data, and environment variables.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.service = os.getenv("SERVICE_NAME", "unknown_service")
        self.environment = os.getenv("ENVIRONMENT", "unknown_env")

    def format(self, record):
        log_record = {
            "level": record.levelname,
            "time": self.formatTime(record, self.datefmt),
            "logger": record.name,  # Include logger name
            "module": record.module,
            "message": self._format_message(record),
            "pathname": record.pathname,
            "lineno": record.lineno,
            "process": record.process,
            "thread": record.thread,
            "service": self.service,
            "environment": self.environment,
        }

        for attr, value in record.__dict__.items():
            if attr not in log_record and not attr.startswith("_"):
                log_record[attr] = value

        return json.dumps(log_record)

    def _format_message(self, record):
        """
        Handle the case where arguments are passed to the logger (like in logger.info('%s %s', arg1, arg2)).
        This method ensures that the message is correctly formatted.
        """
        if record.args:
            return record.getMessage() % record.args
        return record.getMessage()



from logging import getLogger

import requests
from django.conf import settings

logger = getLogger(__file__)

api_key = settings.AI_SERVICE_API_KEY
base_url = settings.AI_SERVICE_DEMO_URL
headers = {"api-key": api_key}


class AIProcessing:
    """
    This class handles the AI processing for submissions

    For this to work, you must specify the AI_SERVICE_DEMO_URL
    and LEARNGUAL_AI_API_KEY in your settings.py file.
    """

    @staticmethod
    def analyze_audio(
        audio,
        reference_text,
        scripted: bool = False,
        prompt: str = None,
        language: str = None,
        query_string: str = None,
    ):
        """
        this communicates with the AI model and returns the analysis results
        for the given audio
        """
        base_url = settings.AI_SERVICE_DEMO_URL
        if not str(base_url).endswith("/"):
            base_url += "/"

        files = {"audio_data": ("audio.mp3", audio)}
        payload = {"reference_text": reference_text, "scripted": scripted}

        if prompt:
            payload["prompt"] = prompt

        if language:
            payload["language"] = language

        if query_string:
            base_url += "?" + query_string

        response = requests.post(
            url=base_url,
            headers=headers,
            files=files,
            data=payload,
        )
        if response.status_code == 200:
            return {"status": True, "response": response}

        else:
            return {"status": False, "response": response}

    @staticmethod
    def relevance(topic: str, essay: str, query_string: str = None):
        base_url = settings.AI_SERVICE_DEMO_URL

        if not str(base_url).endswith("/"):
            base_url += "/"

        base_url = base_url.rstrip("process/") + "/relevance/"

        if query_string:
            base_url += "?" + query_string

        response = requests.post(
            url=base_url,
            headers=headers,
            data={"topic": topic, "essay": essay},
        )

        if response.status_code == 200:
            return {"status": True, "response": response}

        else:
            return {"status": False, "response": response}

    def grammar(text_body: str, query_string: str = None):
        base_url: str = settings.AI_SERVICE_DEMO_URL

        if not str(base_url).endswith("/"):
            base_url += "/"

        base_url = base_url.rstrip("process/") + "/gammar/"
        if query_string:
            base_url += "?" + query_string

        response = requests.post(
            url=base_url,
            headers=headers,
            data={"speech": text_body},
        )

        if response.status_code == 200:
            return {"status": True, "response": response}

        else:
            return {"status": False, "response": response}



import time
import zoneinfo
from logging import getLogger

import pytz
from django.utils import timezone, translation

from .utils import get_language_code

logger = getLogger(__file__)


class TimeZoneMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        tz_header = request.headers.get("TZ")

        if tz_header:
            try:
                timezone.activate(tz_header)
            except (pytz.UnknownTimeZoneError, zoneinfo.ZoneInfoNotFoundError):
                logger.error("Invalid timezone %s", tz_header)
                pass  # Handle unknown timezone error here
        else:
            # Set default timezone if TZ header is not provided
            timezone.activate("UTC")

        response = self.get_response(request)
        timezone.deactivate()
        return response


class LanguageMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Extract language from _lang query parameter
        lang = request.GET.get("_lang")
        if lang:
            lang = get_language_code(lang).upper()
        if lang:
            # Activate the new language if it's valid
            translation.activate(lang)
        else:
            # Fallback to default language if not valid
            translation.activate("EN")

        response = self.get_response(request)
        # Restore the original language
        translation.activate("EN")
        return response


class RequestTimeMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Record the start time
        start_time = time.time()

        # Process the request
        response = self.get_response(request)

        # Calculate the time taken
        duration = time.time() - start_time
        logger.info(f"Request to {request.path} took {duration:.4f} seconds")

        response["X-Request-Duration"] = f"{duration:.4f} seconds"
        return response
